/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL1_Abs_Corres_2  (Level 1 library routine)                        */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    17 JUL 94   Version 1.0                                      */
/*              02 NOV 95   Version 2.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure determines the set of pairs of absolute alignment offsets for */
/* a pair of positions, given by some origin and offsets with respect to it; */
/* structures containing data about the origins and the alignment are        */
/* provided as input parameters; returns the standard error structure, NULL  */
/* if there were no errors                                                   */
/*                                                                           */
/* NOTE: If there is no absolute interval correspondent to the given         */
/*       interval, an empty (NULL) list is returned as the result, but the   */
/*       error status remains unaffected (NULL); however, if the sequence    */
/*       containing the origin with respect to which the interval is given   */
/*       is not present in the alignment, an appropriate error status is     */
/*       returned                                                            */
/*                                                                           */
/*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl1.h"

/*****************************************************************************/
/*                                                                           */
/* Definitions section                                                       */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/* Definitions of local constants of the unit                                */
/*****************************************************************************/


/* Limits for predefined buffer sizes                                        */

#define ERR_MSGLIMIT             128


/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/


errind NTL1_AC2_Assemble_Error (int severity, int code,
                                char *comment, int description);


/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/


static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* PROCEDURE: NTL1_Abs_Corres_2                                              */
/*                                                                           */
/* Central procedure for the unit; returns NULL if the absolute positions    */
/*   have been found and set of such pairs allocated, error diagnostics      */
/*   structure otherwise                                                     */
/* NOTE: If the corresponding absolute positions are not found, but          */
/*       otherwise there were no errors, NULL is still returned, however the */
/*       list of results is empty                                            */

errind NTL1_Abs_Corres_2 (char *reference,
                          long int offset_1, long int offset_2,
                          orgref_ptr origins,
                          header_ptr file_data, align_ptr alignment,
                          twoint_ptr *result)
{
 char *owner_seq; long int reference_position; orgref_ptr or;
 long int abs_seq_1, abs_seq_2, swap_abs, lookup_1, lookup_2, counter, scanner;
 int index, trunc; bool flipped, found_1, found_2;
 line_ptr tracker; twoint_ptr new_corr;
 
 /* Check the legality of provided offsets with respect to origin position   */

 if ((offset_1 == 0) || (offset_2 == 0))
  return NTL1_AC2_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                  "Zero offsets not permitted", 0);

 /* Determine the "owner" sequence for the given origin (reference) site     */

 owner_seq = NULL; reference_position = 0;
 for (or = origins; (or != NULL) && (owner_seq == NULL); or = or -> next) {
  if (!strcmp (or -> site_name, reference)) {
   owner_seq = NTL0_strsave (or -> in_species);
   reference_position = or -> location;
  }
 }
 if (owner_seq == NULL) {
  sprintf (Error_Message, "Site <%s> not found in origins file", reference);
  return NTL1_AC2_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                  Error_Message, 1);
 }
 /* Check whether the "owner" sequence is present in the alignment at all    */

 found_1 = FALSE; index = 0;
 while ((!found_1) && (index < file_data -> dimension)) {
  if (!strcmp (((file_data -> sequences) [index]).seq_name, owner_seq))
   found_1 = TRUE;
  else index++;
 }
 if (!found_1) {
  sprintf (Error_Message,
           "Sequence <%s> containing origin %s not in the alignment",
           owner_seq, reference);
  return NTL1_AC2_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                  Error_Message, 2);
 }
 /* Calculate the absolute position of the reference in its sequence         */

 if ((reference_position == 1) && (offset_1 < 0)) abs_seq_1 = offset_1;
 else if (offset_1 < 0) abs_seq_1 = reference_position + offset_1;
 else abs_seq_1 = reference_position + offset_1 - 1; /* Compensate lack of 0 */
 if ((reference_position == 1) && (offset_2 < 0)) abs_seq_2 = offset_2;
 else if (offset_2 < 0) abs_seq_2 = reference_position + offset_2;
 else abs_seq_2 = reference_position + offset_2 - 1; /* Compensate lack of 0 */

 *result = NULL;      /* Start with an empty list of absolute correspondents */
 
 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
 /* Check the legality of calculated absolute positions                      */
 /*                                                                          */
 /* if ((abs_seq_1 <= 0) && (abs_seq_2 <= 0))                                */
 /*  return NTL1_AC2_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,          */
 /*                                  "Illegal offset out of alignment", 0);  */
 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

 if ((abs_seq_1 <= 0) && (abs_seq_2 <= 0)) return NULL;  /* Not in alignment */
 
 if (abs_seq_1 > abs_seq_2) {
  flipped = TRUE;
  swap_abs = abs_seq_1; abs_seq_1 = abs_seq_2; abs_seq_2 = swap_abs;
 }
 else flipped = FALSE;      /* Interval provided in correct (low,high) order */

 /* Locate the absolute positions in the alignment for both critical sites   */

 for (index = 0; index < file_data -> dimension; index++) {
  if (!strcmp (((file_data -> sequences) [index]).seq_name, owner_seq)) {

   /* Examine the endpoints of the given interval for possible truncation    */
   
   if ((abs_seq_1 >= (alignment -> begin) [index]) &&
       (abs_seq_1 <= (alignment -> end) [index]) &&
       (abs_seq_2 >= (alignment -> begin) [index]) &&
       (abs_seq_2 <= (alignment -> end) [index])) {  /* Completely contained */
       
    lookup_1 = abs_seq_1; lookup_2 = abs_seq_2; trunc = 0;
   }
   else if ((abs_seq_1 >= (alignment -> begin) [index]) &&
            (abs_seq_1 <= (alignment -> end) [index])) {   /* Beyond the end */
            
    lookup_1 = abs_seq_1; lookup_2 = (alignment -> end) [index];
    trunc = RIGHT_TRUNC_CODE;
   }
   else if ((abs_seq_2 >= (alignment -> begin) [index]) &&
            (abs_seq_2 <= (alignment -> end) [index])) { /* Before beginning */
            
    lookup_1 = (alignment -> begin) [index]; lookup_2 = abs_seq_2;
    trunc = LEFT_TRUNC_CODE;
   }
   else if ((abs_seq_1 < (alignment -> begin) [index]) &&
            (abs_seq_2 > (alignment -> end) [index])) {   /* Containing seq. */
            
    lookup_1 = (alignment -> begin) [index];
    lookup_2 = (alignment -> end) [index];
    trunc = BOTH_TRUNC_CODE; if (flipped) trunc = -trunc;
   }
   else {  lookup_1 = lookup_2 = -1; }           /* No overlap with sequence */

   if ((lookup_1 != -1) && (lookup_2 != -1)) {    /* If there is overlap ... */

    /* Locate a record for the new absolute interval in the correspondents   */

    new_corr = (twoint_ptr) NTL0_ckalloc (sizeof (TwoInt_Struct));
    new_corr -> trunc = trunc;
    
    counter = 0; scanner = (alignment -> begin) [index];
    found_1 = FALSE; found_2 = FALSE;
    for (tracker = (alignment -> lines) [index];
         (tracker != NULL) && ((!found_1) || (!found_2));
         tracker = tracker -> next) {
     if (tracker -> code == GAP_STRETCH) counter += tracker -> length;
     else if (tracker -> code == TEXT_STRETCH) {
      if (!found_1) {
       if (scanner + tracker -> length > lookup_1) {
        new_corr -> value_1 = counter + lookup_1 - scanner; found_1 = TRUE;
       }
      }
      if (!found_2) {
       if (scanner + tracker -> length > lookup_2) {
        new_corr -> value_2 = counter + lookup_2 - scanner; found_2 = TRUE;
       }
      }
      counter += tracker -> length; scanner += tracker -> length;
     }
    }
    if ((!found_1) || (!found_2))
     return NTL1_AC2_Assemble_Error (FATAL_ERROR, ERR_CODE_PROBLEM,
                                     "Absolute locations not found", 3);
    else {
     if (flipped) {             /* Return to the original order of endpoints */
      swap_abs = new_corr -> value_1;
      new_corr -> value_1 = new_corr -> value_2;
      new_corr -> value_2 = swap_abs;
     }
     new_corr -> next = *result; *result = new_corr;
    }
   }
  }
 }
 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
 /* if (*result == NULL) {  There was no overlap with any corresponding seq. */
 /*  sprintf (Error_Message,                                                 */
 /*           "Referenced positions <%s,%ld,%ld> out of alignment",          */
 /*           reference, offset_1, offset_2);                                */
 /* return NTL1_AC2_Assemble_Error (WARNING, ERR_NO_VALUE, Error_Message, 0);*/
 /* }                                                                        */
 /* else return NULL; */  /* At least one absolute interval has been located */
 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

 return NULL;        /* Corresponding interval(s) found or not - no error(s) */
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_AC2_Assemble_Error                                        */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL1_AC2_Assemble_Error (int severity, int code,
                                char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
                     (strlen (comment) + strlen ("_Abs_Corres_2: ") + 1) *
                                                                sizeof (char));
 sprintf (report, "_Abs_Corres_2: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

