/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL1_Corresponds  (Level 1 library routine)                         */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    08 SEP 94   Version 1.0                                      */
/*              02 NOV 95   Version 2.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure determines new offset (relative to the new origin) of a       */
/* position, given by some origin and offset with respect to it, using the   */
/* alignment file, if the two given origins correspond to different species; */
/* structures containing data about the origins and the alignment are        */
/* provided as input parameters; returns the standard error structure, NULL  */
/* if there were no errors                                                   */
/*                                                                           */
/* NOTE: If sequences containing both origins (old and new) are present in   */
/*       the alignment, but there is no correspondent to the old position,   */
/*       an empty (NULL) list of results is returned, without reporting an   */
/*       error                                                               */
/*                                                                           */
/*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl1.h"

/*****************************************************************************/
/*                                                                           */
/* Definitions section                                                       */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/* Definitions of local constants of the unit                                */
/*****************************************************************************/


/* Limits for predefined buffer sizes                                        */

#define ERR_MSGLIMIT             128


/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/


errind NTL1_CD_Assemble_Error (int severity, int code,
                                char *comment, int description);


/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/


static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* PROCEDURE: NTL1_Corresponds                                               */
/*                                                                           */
/* Central procedure for the unit; returns NULL if there were no errors in   */
/*   determining the corresponding position, error diagnostics structure     */
/*   otherwise                                                               */

errind NTL1_Corresponds (char *reference, long int offset, char *new_origin,
                         orgref_ptr origins, header_ptr file_data,
                         align_ptr alignment, appint_ptr *result)
{
 char *old_owner_seq, *new_owner_seq; long int old_pos, new_pos;
 long int abs_oldseq, al_position, counter, scanner, at_abs, corresponding;
 orgref_ptr org; line_ptr tracker;
 int count_found, index; bool found;
 long int trail_align, trail_seq, next_align, next_seq,
          trailer_absolute, trailer_offset;
 appint_ptr transfer, new_lead, new_hit, trash;

 /* Check the legality of the provided old offset                            */

 if (offset == 0) {
  return NTL1_CD_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 "Zero offsets not permitted", 1);
 }
 
 /* Determine the "owner" sequnce(s) for "old" and "new" origin site         */

 old_owner_seq = new_owner_seq = NULL; old_pos = new_pos = 0;
 for (org = origins;
      (org != NULL) && ((old_owner_seq == NULL) || (new_owner_seq == NULL));
      org = org -> next) {
  if (!strcmp (org -> site_name, reference)) {
   old_owner_seq = NTL0_strsave (org -> in_species); old_pos = org -> location;
  }
  if (!strcmp (org -> site_name, new_origin)) {
   new_owner_seq = NTL0_strsave (org -> in_species); new_pos = org -> location;
  }
 }
 if (old_owner_seq == NULL) {
  sprintf (Error_Message, "Site <%s> not found in origins file", reference);
  return NTL1_CD_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 Error_Message, 2);
 }
 else if (new_owner_seq == NULL) {
  sprintf (Error_Message, "Site <%s> not found in origins file", new_origin);
  return NTL1_CD_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 Error_Message, 3);
 }
 /* Check whether the "owner" sequence of the old origin is in the alignment */

 found = FALSE; index = 0;
 while ((!found) && (index < file_data -> dimension)) {
  if (!strcmp (((file_data -> sequences) [index]).seq_name, old_owner_seq))
   found = TRUE;
  else index++;
 }
 if (!found) {
  sprintf (Error_Message,
           "Sequence <%s> containing the old origin <%s> not in the alignment",
           old_owner_seq, reference);
  return NTL1_CD_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 Error_Message, 4);
 }
 /* Check whether the "owner" sequence of the new origin is in the alignment */

 found = FALSE; index = 0;
 while ((!found) && (index < file_data -> dimension)) {
  if (!strcmp (((file_data -> sequences) [index]).seq_name, new_owner_seq))
   found = TRUE;
  else index++;
 }
 if (!found) {
  sprintf (Error_Message,
           "Sequence <%s> containing the new origin <%s> not in the alignment",
           new_owner_seq, new_origin);
  return NTL1_CD_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,
                                 Error_Message, 5);
 }
 *result = NULL;               /* Start with an empty list of correspondents */

 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
 /* Check the legality of the provided position for conversion               */
 /*                                                                          */
 /* if ((offset < 0) && (old_pos + offset <= 0)) {                           */
 /*  return NTL1_CD_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,           */
 /*                            "Position for conversion out of sequence", 4);*/
 /* }                                                                        */
 /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

 /* Check if both old and new origin pertain to the same sequence            */

 if (!strcmp (old_owner_seq, new_owner_seq)) {  /* Both origins in same seq. */
  *result = (appint_ptr) NTL0_ckalloc (sizeof (AppInt_Struct));
  (*result) -> value = old_pos - new_pos + offset;
  if ((offset > 0) && ((*result) -> value <= 0)) ((*result) -> value)--;
  else if ((offset < 0) && ((*result) -> value >= 0)) ((*result) -> value)++;
  (*result) -> correction = 0; (*result) -> next = NULL;
  return NULL;
 }
 else {         /* Given origins correspond to different species (sequences) */

  /* Calculate the absolute position of the reference in its sequence        */

  if ((offset < 0) && (old_pos == 1)) abs_oldseq = offset;
  else if (offset < 0) abs_oldseq = old_pos + offset;
  else abs_oldseq = old_pos + offset - 1; /* Compensate for lack of 0 offset */

  if (abs_oldseq < 0) return NULL;       /* Negative ranges not in alignment */
  
  /* Locate the absolute position in the alignment for the critical site     */
  /* Check whether the absolute position is unique within the alignment      */

  count_found = 0; al_position = 0; transfer = NULL;
  for (index = 0; index < file_data -> dimension; index++) {
   if (!strcmp (((file_data -> sequences) [index]).seq_name, old_owner_seq)) {
    if ((abs_oldseq >= (alignment -> begin) [index]) &&
        (abs_oldseq <= (alignment -> end) [index])) {
     new_lead = (appint_ptr) NTL0_ckalloc (sizeof (AppInt_Struct));
     counter = 0; scanner = (alignment -> begin) [index]; found = FALSE;
     for (tracker = (alignment -> lines) [index];
          (tracker != NULL) && (scanner <= abs_oldseq);
          tracker = tracker ->next) {
      if (tracker -> code == GAP_STRETCH) counter += tracker -> length;
      else if (tracker -> code == TEXT_STRETCH) {
       if (scanner + tracker -> length > abs_oldseq) {
        count_found++; al_position = counter + abs_oldseq - scanner;
        new_lead -> value = al_position;
        new_lead -> next = transfer; transfer = new_lead;
        found = TRUE;
       }
       else counter += tracker -> length;
       scanner += tracker -> length;
      }
     }
     if (!found) free (new_lead);
    }
   }
  }
  if (count_found == 0) return NULL; /* Referenced position out of alignment */

  /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
  /* if (count_found != 1) {                                                 */
  /*  if (count_found == 0) {                                                */
  /*   sprintf (Error_Message,                                               */
  /*            "Referenced position <%s,%ld> out of alignment",             */
  /*            reference, offset);                                          */
  /*   return NTL1_CD_Assemble_Error (USER_ERROR, ERR_BAD_REFERENCE,         */
  /*                                  Error_Message, 5);                     */
  /*  }                                                                      */
  /* }                                                                       */
  /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
  
  /* Locate the target sequence in alignment that has character at position  */
  /* Check whether the located target sequence is unique within alignment    */

  new_lead = transfer; while (new_lead != NULL) {
   al_position = new_lead -> value;
   count_found = 0; at_abs = 0; trailer_absolute = -1; trailer_offset = 0;
   for (index = 0; index < file_data -> dimension; index++) {
    if (!strcmp (((file_data -> sequences) [index]).seq_name, new_owner_seq)) {
     trail_align = -1; trail_seq = (alignment -> begin) [index] - 1;
     counter = 0; corresponding = (alignment -> begin) [index];
     for (tracker = (alignment -> lines) [index];
          (tracker != NULL) && (counter <= al_position);
          tracker = tracker ->next) {
      if (tracker -> code == GAP_STRETCH) counter += tracker -> length;
      else if (tracker -> code == TEXT_STRETCH) {
       if (counter + tracker -> length > al_position) {
        count_found++; at_abs = corresponding + al_position - counter;
        new_hit = (appint_ptr) NTL0_ckalloc (sizeof (AppInt_Struct));
        if (at_abs >= new_pos) new_hit -> value = at_abs - new_pos + 1;
        else new_hit -> value = at_abs - new_pos;
        new_hit -> correction = 0;
        new_hit -> next = *result; *result = new_hit;
       }
       else {
        corresponding += tracker -> length;

        /* Keep track of position of the end of the last strip of text seen  */
        /*   in this sequence, in the alignment and in seq. absolute pos.    */

        trail_align = counter + tracker -> length - 1;
        trail_seq = trail_seq + tracker -> length;
       }
       counter += tracker -> length;
      }
     }
     /* If no corresponding position has been found yet, find the beginning  */
     /*   of the next text piece in the sequence (if any), its offset from   */
     /*   the beginning of the alignment and its position within its sequence*/
     /* Next, compare the recorded value of the smallest distance between the*/
     /*   neighbouring edges and the position with new distances and update  */
     /*   values, if necessary: value of the absolute alignment position of  */
     /*   the nearest neighbouring edge and its offset within its sequence   */

     if (count_found == 0) {
      if (tracker == NULL) { next_align = -1; next_seq = -1; }
      else if (counter < alignment -> size) {
       next_align = counter;
       if (trail_align == -1) next_seq = (alignment -> begin) [index];
       else next_seq = trail_seq + 1;
      }
      if (trail_align != -1) {
       if (next_align != -1) {
        if (al_position - trail_align > next_align - al_position) {
         if (trailer_absolute < 0) {
          trailer_absolute = next_align; trailer_offset = next_seq;
         }
         else {
          if (NTL0_labs (al_position - trailer_absolute) >
              next_align - al_position) {
           trailer_absolute = next_align; trailer_offset = next_seq;
          }
         }
        }
        else {
         if (trailer_absolute < 0) {
          trailer_absolute = trail_align; trailer_offset = trail_seq;
         }
         else {
          if (NTL0_labs (al_position - trailer_absolute) >
              al_position - trail_align) {
           trailer_absolute = trail_align; trailer_offset = trail_seq;
          }
         }
        } 
       }
       else {          /* There is no next neighbouring position in alignment */
        if (trailer_absolute < 0) {
         trailer_absolute = trail_align; trailer_offset = trail_seq;
        }
        else {
         if (NTL0_labs (al_position - trailer_absolute) >
             al_position - trail_align) {
          trailer_absolute = trail_align; trailer_offset = trail_seq;
         }
        }
       }
      }
      else {
       if (next_align != -1) {
        if (trailer_absolute < 0) {
         trailer_absolute = next_align; trailer_offset = next_seq;
        }
        else {
         if (NTL0_labs (al_position - trailer_absolute) >
             next_align - al_position) {
          trailer_absolute = next_align; trailer_offset = next_seq;
         }
        }
       }
      }
     }
    }
   }
   if (count_found == 0) {
    if (trailer_absolute >= 0) {
     /* Although there was no direct correspondent to the given position in  */
     /*   the alignment, there was a neighbouring position found; return the */
     /*   neighbours offset and the size of the gap dividing the "right"     */
     /*   corresponding position and the selected neighbour                  */

     new_hit = (appint_ptr) NTL0_ckalloc (sizeof (AppInt_Struct));
     if (trailer_offset >= new_pos)
      new_hit -> value = trailer_offset - new_pos + 1;
     else new_hit -> value = trailer_offset - new_pos;
     if (al_position < trailer_absolute)
      new_hit -> correction = trailer_absolute - al_position;
     else new_hit -> correction = trailer_absolute - al_position;
     new_hit -> next = *result; *result = new_hit;
    }
   }
   trash = new_lead; new_lead = new_lead -> next; free (trash);
  }

  /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */
  /* if (*result == NULL) {                                                  */
  /*  sprintf (Error_Message,                                                */
  /*           "No match for <%s,%ld> in alignment, with respect to <%s>",   */
  /*           reference, offset, new_origin);                               */
  /*  return NTL1_CD_Assemble_Error (USER_ERROR, ERR_NO_VALUE,               */
  /*                                 Error_Message, 6);                      */
  /* }                                                                       */
  /* else return NULL;                                                       */
  /* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

  return NULL;    /* No error reported, regardless of whether results filled */
 }
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_CD_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL1_CD_Assemble_Error (int severity, int code,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
                     (strlen (comment) + strlen ("_Corresponds: ") + 1) *
                                                                sizeof (char));
 sprintf (report, "_Corresponds: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

