/*****************************************************************************/
/*                                                                           */
/* UNIT: NTL1_Compress_Plain (Level 1 library routine)                       */
/*                                                                           */
/* Author: Nikola Stojanovic                                                 */
/*                                                                           */
/* Revision:    08 APR 97   Version 1.0                                      */
/*                                                                           */
/* Function:                                                                 */
/*                                                                           */
/*   Procedure receives a list of records loaded from plain form landmark    */
/* files, sorted in the increasing order of the starting positions and makes */
/* a list of overlaps encountered within the records; returns the error      */
/* structure if there were any errors encountered, NULL otherwise            */
/*                                                                           */
/* NOTE: Procedure is desctructive for the received list of records!         */
/*                                                                           */
/*****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ntl1.h"

/*****************************************************************************/
/*                                                                           */
/* Definitions section                                                       */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/* Definitions of local constants of the unit                                */
/*****************************************************************************/


/* Limit for the length of an error message that can be assembled            */

#define ERR_MSGLIMIT             128


/*****************************************************************************/
/* Prototypes of all locally used functions of this unit                     */
/*****************************************************************************/


errind NTL1_CP_Consensus_Text (char **consensus, char *add_string,
                               long int position);
errind NTL1_CP_Merge_Characters (char ch1, char ch2, char *result);
errind NTL1_CP_Assemble_Error (int severity, int code,
                               char *comment, int description);


/*****************************************************************************/
/* Definitions of global (static) variables of the unit                      */
/*****************************************************************************/


static char Error_Message [ERR_MSGLIMIT]; /* Temporary buffer, error passing */

/*****************************************************************************/
/*                                                                           */
/* Code section                                                              */
/*                                                                           */
/*****************************************************************************/


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_Compress_Plain                                            */
/*                                                                           */
/* Main (interface) procedure of this unit                                   */

errind NTL1_Compress_Plain (plain_ptr *list, plist_ptr *compressed)
{
 errind erret; int item_count, start_count, end_position; char *consensus;
 plist_ptr last_compressed, new_compressed; plain_ptr item, sublist, scan;

 *compressed = NULL;                           /* Initialize the output list */
 item_count = 0; last_compressed = NULL;

 /* Loop for all records in the input list, until the list gets empty        */

 while (*list != NULL) {
 
  /* Get the first item in the sublist of possible overlaps - prepare for it */
  
  item = *list; *list = (*list) -> next; item -> next = NULL; item_count++;
  sublist = item; end_position = item -> stop; start_count = item_count;
  consensus = NTL0_strsave (item -> text);

  /* Loop through the original list as long as there are overlaps - merge    */
  
  while ((*list != NULL) && ((*list) -> start <= end_position)) {
  
   item = *list; *list = (*list) -> next; item -> next = NULL; item_count++;
   scan = sublist; while (scan -> next != NULL) scan = scan -> next;
   scan -> next = item;

   /* Adjust the end of overlap area and determine the consensus for it      */
   
   if (item -> stop > end_position) end_position = item -> stop;
   if ((erret = NTL1_CP_Consensus_Text (&consensus, item -> text,
                 (item -> start) - (sublist -> start))) != NULL) return erret;
  }
  /* After all overlaps are isolated, create the header for it and connect   */
  
  new_compressed = (plist_ptr) NTL0_ckalloc (sizeof (Plist_Struct));
  
  new_compressed -> lines = sublist;
  new_compressed -> from = sublist -> start;
  new_compressed -> to = end_position;
  new_compressed -> start_count = (int) start_count;
  new_compressed -> stop_count = (int) item_count;
  new_compressed -> consensus = consensus;
  new_compressed -> next = NULL;
  
  if (last_compressed == NULL) *compressed = new_compressed;
  else last_compressed -> next = new_compressed;
  last_compressed = new_compressed;
 }
 return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_CP_Consensus_Text                                         */
/*                                                                           */
/* Procedure receives two overlapping strings and determines the consensus   */
/*   in the area of overlap, adjusting the first received to the whole       */
/*   length and contents of the overlap; returns the error structure, if any */
/*   errors have been encountered, NULL otherwise                            */

errind NTL1_CP_Consensus_Text (char **consensus, char *add_string,
                               long int position)
{
 long int base_length, add_length, new_length, index; char *new_cons, ch;
 errind erret;

 /* Determine the proper length of the consensus for the union of strings    */
 
 base_length = strlen (*consensus); add_length = strlen (add_string);
 if (position > base_length - 1)    /* New string must start overlapping old */
  return NTL1_CP_Assemble_Error (SYSTEM_ERROR, ERR_BAD_VALUE,
                                 "Bad position calculation", 0);
 if (position + add_length < base_length) new_length = base_length + 1;
 else new_length = position + add_length + 1;

 /* Allocate space for the combined length of the overlap and flanking parts */
 
 new_cons = (char *) NTL0_ckalloc (new_length * sizeof (char));

 strcpy (new_cons, *consensus);  /* Start same as old, unless adjusted below */
 
 for (index = position; index < position + add_length; index++) {
  if (index < base_length) {          /* If still in the area of the overlap */
  
   if (add_string [index - position] != (*consensus) [index]) {
    if ((erret = NTL1_CP_Merge_Characters (add_string [index - position],
					  (*consensus) [index], &ch)) != NULL)
     return erret;
    new_cons [index] = ch;   /* Assign the consensus character to the string */
   }
  }
  else new_cons [index] = add_string [index - position];    /* Just copy ... */
 }
 /* Conclude the consensus string and make the output parameter assignment   */
 
 new_cons [new_length - 1] = '\0'; free (*consensus); *consensus = new_cons;
 return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_CP_Merge_Characters                                       */
/*                                                                           */
/* Procedure receives two characters, and determines their consensus         */
/*   character, in accordance with the convention for the letters in DNA     */
/*   sequences; returns the error structure if some of the letters could not */
/*   be interpreted, NULL otherwise                                          */

errind NTL1_CP_Merge_Characters (char ch1, char ch2, char *result)
{
 if (ch1 == ch2) *result = ch1;       /* If characters are same, just return */

 /* If any of the characters is a gap symbol, then consensus must be gap     */
 
 else if ((ch1 == GAP_SYMBOL) || (ch2 == GAP_SYMBOL)) *result = GAP_SYMBOL;

 /* If any character is "just anything", it subsumes any other symbol        */
 
 else if ((ch1 == 'N') || (ch1 == 'X') || (ch2 == 'N') || (ch2 == 'X'))
  *result = 'N';

 /* If there is genuine choice to be made, do the case analysis              */
 
 else if (ch1 == 'A') {
  if (ch2 == 'C') *result = 'M';
  else if (ch2 == 'G') *result = 'R';
  else if (ch2 == 'T') *result = 'W';
  else if ((ch2 == 'W') || (ch2 == 'R') || (ch2 == 'M') ||
           (ch2 == 'D') || (ch2 == 'H') || (ch2 == 'V')) *result = ch2;
  else if (ch2 == 'Y') *result = 'H';
  else if (ch2 == 'S') *result = 'V';
  else if (ch2 == 'K') *result = 'D';
  else if (ch2 == 'B') *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'C') {
  if (ch2 == 'A') *result = 'M';
  else if (ch2 == 'G') *result = 'S';
  else if (ch2 == 'T') *result = 'Y';
  else if ((ch2 == 'Y') || (ch2 == 'S') || (ch2 == 'M') ||
           (ch2 == 'B') || (ch2 == 'H') || (ch2 == 'V')) *result = ch2;
  else if (ch2 == 'W') *result = 'H';
  else if (ch2 == 'R') *result = 'V';
  else if (ch2 == 'K') *result = 'B';
  else if (ch2 == 'D') *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'G') {
  if (ch2 == 'A') *result = 'R';
  else if (ch2 == 'C') *result = 'S';
  else if (ch2 == 'T') *result = 'K';
  else if ((ch2 == 'S') || (ch2 == 'R') || (ch2 == 'K') ||
           (ch2 == 'B') || (ch2 == 'D') || (ch2 == 'V')) *result = ch2;
  else if (ch2 == 'W') *result = 'D';
  else if (ch2 == 'Y') *result = 'B';
  else if (ch2 == 'M') *result = 'V';
  else if (ch2 == 'H') *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'T') {
  if (ch2 == 'A') *result = 'W';
  else if (ch2 == 'C') *result = 'Y';
  else if (ch2 == 'G') *result = 'K';
  else if ((ch2 == 'W') || (ch2 == 'Y') || (ch2 == 'K') ||
           (ch2 == 'B') || (ch2 == 'D') || (ch2 == 'H')) *result = ch2;
  else if (ch2 == 'S') *result = 'B';
  else if (ch2 == 'R') *result = 'D';
  else if (ch2 == 'M') *result = 'H';
  else if (ch2 == 'V') *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'W') {
  if ((ch2 == 'A') || (ch2 == 'T')) *result = ch1;
  else if ((ch2 == 'D') || (ch2 == 'H')) *result = ch2;
  else if ((ch2 == 'S') || (ch2 == 'B') || (ch2 == 'V')) *result = 'N';
  else if (ch2 == 'C') *result = 'H';
  else if (ch2 == 'G') *result = 'D';
  else if (ch2 == 'Y') *result = 'H';
  else if (ch2 == 'R') *result = 'D';
  else if (ch2 == 'M') *result = 'H';
  else if (ch2 == 'K') *result = 'D';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'Y') {
  if ((ch2 == 'C') || (ch2 == 'T')) *result = ch1;
  else if ((ch2 == 'B') || (ch2 == 'H')) *result = ch2;
  else if ((ch2 == 'R') || (ch2 == 'D') || (ch2 == 'V')) *result = 'N';
  else if (ch2 == 'A') *result = 'H';
  else if (ch2 == 'G') *result = 'B';
  else if (ch2 == 'W') *result = 'H';
  else if (ch2 == 'S') *result = 'B';
  else if (ch2 == 'M') *result = 'H';
  else if (ch2 == 'K') *result = 'B';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'S') {
  if ((ch2 == 'C') || (ch2 == 'G')) *result = ch1;
  else if ((ch2 == 'B') || (ch2 == 'V')) *result = ch2;
  else if ((ch2 == 'W') || (ch2 == 'D') || (ch2 == 'H')) *result = 'N';
  else if (ch2 == 'A') *result = 'V';
  else if (ch2 == 'T') *result = 'B';
  else if (ch2 == 'Y') *result = 'B';
  else if (ch2 == 'R') *result = 'V';
  else if (ch2 == 'M') *result = 'V';
  else if (ch2 == 'K') *result = 'B';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'R') {
  if ((ch2 == 'A') || (ch2 == 'G')) *result = ch1;
  else if ((ch2 == 'D') || (ch2 == 'V')) *result = ch2;
  else if ((ch2 == 'Y') || (ch2 == 'B') || (ch2 == 'H')) *result = 'N';
  else if (ch2 == 'C') *result = 'V';
  else if (ch2 == 'T') *result = 'D';
  else if (ch2 == 'W') *result = 'D';
  else if (ch2 == 'S') *result = 'V';
  else if (ch2 == 'M') *result = 'V';
  else if (ch2 == 'K') *result = 'D';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'M') {
  if ((ch2 == 'A') || (ch2 == 'C')) *result = ch1;
  else if ((ch2 == 'H') || (ch2 == 'V')) *result = ch2;
  else if ((ch2 == 'K') || (ch2 == 'B') || (ch2 == 'D')) *result = 'N';
  else if (ch2 == 'G') *result = 'V';
  else if (ch2 == 'T') *result = 'H';
  else if (ch2 == 'W') *result = 'H';
  else if (ch2 == 'Y') *result = 'H';
  else if (ch2 == 'S') *result = 'V';
  else if (ch2 == 'R') *result = 'V';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'K') {
  if ((ch2 == 'G') || (ch2 == 'T')) *result = ch1;
  else if ((ch2 == 'B') || (ch2 == 'D')) *result = ch2;
  else if ((ch2 == 'M') || (ch2 == 'H') || (ch2 == 'V')) *result = 'N';
  else if (ch2 == 'A') *result = 'D';
  else if (ch2 == 'C') *result = 'B';
  else if (ch2 == 'W') *result = 'D';
  else if (ch2 == 'Y') *result = 'B';
  else if (ch2 == 'S') *result = 'B';
  else if (ch2 == 'R') *result = 'D';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'B') {
  if ((ch2 == 'C') || (ch2 == 'G') || (ch2 == 'T') ||
      (ch2 == 'Y') || (ch2 == 'S') || (ch2 == 'K')) *result = ch1;
  else if ((ch2 == 'A') || (ch2 == 'W') || (ch2 == 'R') || (ch2 == 'M') ||
           (ch2 == 'D') || (ch2 == 'H') || (ch2 == 'V')) *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'D') {
  if ((ch2 == 'A') || (ch2 == 'G') || (ch2 == 'T') ||
      (ch2 == 'W') || (ch2 == 'R') || (ch2 == 'K')) *result = ch1;
  else if ((ch2 == 'C') || (ch2 == 'Y') || (ch2 == 'S') || (ch2 == 'M') ||
           (ch2 == 'B') || (ch2 == 'H') || (ch2 == 'V')) *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'H') {
  if ((ch2 == 'A') || (ch2 == 'C') || (ch2 == 'T') ||
      (ch2 == 'W') || (ch2 == 'Y') || (ch2 == 'M')) *result = ch1;
  else if ((ch2 == 'G') || (ch2 == 'S') || (ch2 == 'R') || (ch2 == 'K') ||
           (ch2 == 'B') || (ch2 == 'D') || (ch2 == 'V')) *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else if (ch1 == 'V') {
  if ((ch2 == 'A') || (ch2 == 'C') || (ch2 == 'G') ||
      (ch2 == 'S') || (ch2 == 'R') || (ch2 == 'M')) *result = ch1;
  else if ((ch2 == 'T') || (ch2 == 'W') || (ch2 == 'Y') || (ch2 == 'K') ||
           (ch2 == 'B') || (ch2 == 'D') || (ch2 == 'H')) *result = 'N';
  else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                      "Can't merge sequence characters", 0);
 }
 else return NTL1_CP_Assemble_Error (USER_ERROR, ERR_BAD_VALUE,
                                     "Can't merge sequence characters", 0);
 return NULL;
}


/*****************************************************************************/
/*                                                                           */
/* Procedure: NTL1_CP_Assemble_Error                                         */
/*                                                                           */
/* Service procedure for assembling and returnning an error report, based on */
/*   the values of the input parameters; returns the record with the report  */

errind NTL1_CP_Assemble_Error (int severity, int code,
                               char *comment, int description)
{
 char *report; errind assembled;

 report = (char *) NTL0_ckalloc (
   (strlen (comment) + strlen ("_Compress_Plain_List: ") + 1) * sizeof (char));
 sprintf (report, "_Compress_Plain_List: %s", comment);
 assembled = NTL1_Error_Record (severity, code, report, description);
 free (report); return assembled;
}

